package net.sf.jagg;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * This class represents the result of a "group by" operation, where certain
 * fields of a type are selected for a "group by", and certain values can
 * be extracted by referring to <code>Aggregators</code>.
 *
 * @author Randy Gettman
 * @since 0.1.0
 */
public class AggregateValue<T>
{
   private T myObject;
   private Map<Aggregator, Object> myValuesMap;
   private List<Object> myValuesList;

   /**
    * Create an <code>AggregateValue</code> that wraps the given object.  It
    * will also store aggregation values.
    *
    * @param object The object for which this <code>AggregateValue</code> will
    *    wrap.
    */
   public AggregateValue(T object)
   {
      myObject = object;
      myValuesMap = new HashMap<Aggregator, Object>();
      myValuesList = new ArrayList<Object>();
   }

   /**
    * Retrieves the <code>T</code> representing the "group-by" aggregation.
    *
    * @return The <code>T</code> object representing the "group-by" aggregation.
    */
   public T getObject()
   {
      return myObject;
   }

   /**
    * Stores the given <code>value</code> associated with the given
    * <code>Aggregator</code> for later retrieval.  Also appends the given
    * <code>value</code> to an internal list for later retrieval by index.
    *
    * @param agg An <code>Aggregator</code>.
    * @param value The aggregated value.
    * @see Aggregator
    */
   public void setAggregateValue(Aggregator agg, Object value)
   {
      if (myValuesMap.containsKey(agg))
         myValuesMap.remove(agg);
      myValuesMap.put(agg, value);
      myValuesList.add(value);
   }

   /**
    * Retrieves the value for the given <code>Aggregator</code>.
    *
    * @param agg An <code>Aggregator</code>.
    * @return The aggregated value, or <code>null</code> if no such
    *    <code>Aggregator</code> is found.
    */
   public Object getAggregateValue(Aggregator agg)
   {
      return myValuesMap.get(agg);
   }

   /**
    * Retrieves the value for the <code>Aggregator</code> at the given index.
    * @param index The 0-based index.
    * @return The aggregated value.
    * @throws IndexOutOfBoundsException If the index is out of range.
    * @since 0.3.0
    */
   public Object getAggregateValue(int index)
   {
      return myValuesList.get(index);
   }
}
